/*  SQuaLe
 *
 *  Copyright (C) 2004 Julien Moutte <julien@moutte.net>
 *
 *  test-client.c : Source for test client program.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
#include <glib.h>
#include <popt.h>
#include <locale.h>
#include "../lib/squaleclient.h"
#include "../src/squale-i18n.h"

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

static gint *
squale_run_get_column_geometry (SqualeResult *result)
{
  gint *lengths;
  guint i;
  gulong j;
  
  g_return_val_if_fail (result != NULL, NULL);
  
  lengths = g_malloc0 (result->num_fields * sizeof (gint));
  
  /* Column names */
  for (i = 0; i < result->num_fields; i++) {
    lengths[i] = g_utf8_strlen (result->column_names[i], -1);
  }
  
  /* Data */
  for (j = 0; j < result->num_rows; j++) {
    for (i =0; i < result->num_fields; i++) {
      gint length = 0;
      
      if (result->data[j][i]) {
        length= g_utf8_strlen (result->data[j][i], -1);
      }
      if (length > lengths[i]) {
        lengths[i] = length;
      }
    }
  }
  
  return lengths;
}

static void
squale_run_strip_strings (SqualeResult *result)
{
  guint i;
  gulong j;
  
  g_return_if_fail (result != NULL);
  
  /* Column names */
  for (i = 0; i < result->num_fields; i++) {
    if (result->column_names[i]) {
      result->column_names[i] = g_strstrip (result->column_names[i]);
    }
  }
  
  /* Data */
  for (j = 0; j < result->num_rows; j++) {
    for (i =0; i < result->num_fields; i++) {
      if (result->data[j][i]) {
        result->data[j][i] = g_strstrip (result->data[j][i]);
      }
    }
  }
}

int
main (int argc, const char **argv)
{
  gint return_value = 0, timeout = 0, option_id;
  SqualeResult *result = NULL;
  char *socket_name = NULL, *connection_name = NULL, *query = NULL;
  poptContext popt_context;
  struct poptOption options_table[] = 
    {
      POPT_AUTOHELP
      {"socket", 's', POPT_ARG_STRING, &socket_name, 1, N_("Use a specific unix socket (default to /tmp/squale.sock)."), _("SOCKET")},
      {"connection", 'c', POPT_ARG_STRING, &connection_name, 2, N_("Connection which will have to run the query (mandatory)."), _("NAME")},
      {"query", 'q', POPT_ARG_STRING, &query, 3, N_("Query to run against database backend (mandatory)."), _("QUERY")},
      {"timeout", 't', POPT_ARG_INT, &timeout, 4, N_("Maximum delay in seconds before considering the query timed out (default to unlimited)."), _("SECONDS")},
      POPT_TABLEEND
    };

  setlocale (LC_MESSAGES, "");
  setlocale (LC_CTYPE, "");
  bindtextdomain (GETTEXT_PACKAGE, SQUALELOCALEDIR);
  bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");
  textdomain (GETTEXT_PACKAGE);
  
  popt_context = poptGetContext (NULL, argc, argv, options_table, 0);
  
  while ((option_id = poptGetNextOpt (popt_context)) != -1) {
    if (option_id < 1 || option_id > 4) {
      poptPrintHelp (popt_context, stderr, 0);
      return 1;
    }
  }
  
  if ((connection_name == NULL) || (query == NULL)) {
    poptPrintHelp (popt_context, stderr, 0);
    return 1;
  }
  if (socket_name == NULL) {
    socket_name = g_strdup ("/tmp/squale.sock");
  }
  
  poptFreeContext (popt_context);
  
  result = squale_run_query (connection_name, query, socket_name, timeout);
  
  if (!result) {
    g_error (_("An error occured: squale_run_query returned NULL"));
  }
  
  g_message (_("Job was assigned in %dms and processed in %dms"),
             result->assignation_time, result->processing_time);
  
  switch (result->status) {
    case SQUALE_RESULT_ERROR:
      g_message (_("Query generated an error : %s"), result->error);
      return_value = 1;
      break;
    case SQUALE_RESULT_AFFECTED_ROWS:
      g_message (_("Query affected %d rows"), result->affected_rows);
      break;
    case SQUALE_RESULT_RESULTSET_WITH_WARNING:
    case SQUALE_RESULT_RESULTSET:
      {
        guint i;
        gulong j;
        gint *lengths = NULL;
        
        g_message (_("Query generated a resultset with %ld rows " \
                   "and %d columns"), result->num_rows, result->num_fields);
        
        squale_run_strip_strings (result);
        lengths = squale_run_get_column_geometry (result);
        
        /* Column names */
        g_print ("+");
        for (i = 0; i < result->num_fields; i++) {
          for (j = 0; j < lengths[i]; j++) {
            g_print ("-");
          }
          g_print ("+");
        }
        g_print ("\n|");
        for (i = 0; i < result->num_fields; i++) {
          g_print ("%-*s", lengths[i], result->column_names[i]);
          g_print ("|");
        }
        g_print ("\n+");
        for (i = 0; i < result->num_fields; i++) {
          for (j = 0; j < lengths[i]; j++) {
            g_print ("-");
          }
          g_print ("+");
        }
        /* Data */
        for (j = 0; j < result->num_rows; j++) {
          g_print ("\n|");
          for (i =0; i < result->num_fields; i++) {
            g_print ("%-*s", lengths[i], result->data[j][i]);
            g_print ("|");
          }
        }
        g_print ("\n+");
        for (i = 0; i < result->num_fields; i++) {
          for (j = 0; j < lengths[i]; j++) {
            g_print ("-");
          }
          g_print ("+");
        }
        g_print ("\n");
        if (lengths) {
          g_free (lengths);
        }
        if (result->status == SQUALE_RESULT_RESULTSET_WITH_WARNING) {
          g_message (_("Query generated a warning : %s"), result->warning);
        }
        break;
      }
    default:
      g_error (_("An error occured: Unknown result type %d"),
               result->status);
      return_value = 1;
  }
  
  squale_free_result (result);
  
  return (return_value);
}
